/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "SrcInfo.h"
#include "Vector.h"
#include "Map.h"
#include <vector>

using namespace std;
using namespace FD;

class CvfdhHaarDetectObjects;
   
DECLARE_NODE(CvfdhHaarDetectObjects)
/*Node
 *
 * @name CvfdhHaarDetectObjects
 * @category FDOPENCV_HARK:Segmentation
 * @description CvfdhHaarDetectObjects
 *
 * @input_name IMAGEIN
 * @input_type any
 * @input_description Input image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @output_name SRCOUT
 * @output_type Vector<ObjectRef>
 * @output_description Source location. A type of an element of the vector is SrcInfo.
 *
 * @parameter_name SCALE_FACTOR
 * @parameter_type float
 * @parameter_value 1.1
 * @parameter_description The factor by which the search window is scaled between the subsequent scans, for example, 1.1 means increasing window by 10%
 *
 * @parameter_name MIN_NEIGHBORS
 * @parameter_type int
 * @parameter_value 3
 * @parameter_description Minimum number (minus 1) of neighbor rectangles that makes up an object. All the groups of a smaller number of rectangles than min_neighbors-1 are rejected. If min_neighbors is 0, the function does not any grouping at all and returns all the detected candidate rectangles, which may be useful if the user wants to apply a customized grouping procedure
 *
 * @parameter_name FLAGS
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description Mode of operation. Currently the only flag that may be specified is CV_HAAR_DO_CANNY_PRUNING. If it is set, the function uses Canny edge detector to reject some image regions that contain too few or too much edges and thus can not contain the searched object. The particular threshold values are tuned for face detection and in this case the pruning speeds up the processing.
 *
 * @parameter_name MIN_SIZE
 * @parameter_type object
 * @parameter_value <Vector<int> 20 20>
 * @parameter_description Minimum window size (2D vector). By default, it is set to the size of samples the classifier has been trained on (~20x20 for face detection). 
 *
 * @parameter_name MAX_SIZE
 * @parameter_type object
 * @parameter_value <Vector<int> 50 50>
 * @parameter_description Maximum window size (2D vector). This is supported in Opencv after version 2.2. 
 *
 END*/
   
   
class CvfdhHaarDetectObjects : public BufferedNode {

  int m_imageInID;
  int m_srcOutID;
  float scale_factor;
  int min_neighbors;
  int flags;
  Vector<int> min_size;  
  Vector<int> max_size;  
      
public:
  CvfdhHaarDetectObjects(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    m_imageInID = addInput("IMAGEIN");
    m_srcOutID = addOutput("SRCOUT");
    scale_factor = dereference_cast<float>(parameters.get("SCALE_FACTOR"));
    min_neighbors = dereference_cast<int>(parameters.get("MIN_NEIGHBORS"));
    flags = dereference_cast<int>(parameters.get("FLAGS"));
    if (flags != 0) {
      throw new NodeException(this, string("FLAGS is not correctly set."), __FILE__, __LINE__);
    }
    min_size = object_cast<Vector<int> >(parameters.get("MIN_SIZE"));    
    if (min_size.size() != 2) {
      throw new NodeException(this, string("MIN_SIZE is not correctly set."), __FILE__, __LINE__);
    }
    max_size = object_cast<Vector<int> >(parameters.get("MAX_SIZE"));    
    if (max_size.size() != 2) {
      throw new NodeException(this, string("MAX_SIZE is not correctly set."), __FILE__, __LINE__);
    }
    inOrder = true;
  }      
      
  void calculate(int output_id, int count, Buffer &out)
  {

    Vector<ObjectRef>& sources = *new Vector<ObjectRef>;
    (*(outputs[m_srcOutID].buffer))[count] = &sources;

    vector<IplImage*> image_vec;
    image_vec.resize(0);

    IplImage* img_in_tmp;
    ObjectRef in = getInput(m_imageInID, count);
    if (typeid(*in) == typeid(CvfdImage)) {
      RCPtr<CvfdImage> imagePtr = getInput(m_imageInID,count);   
      if(imagePtr->getImage() == NULL){
	return;
      }
      img_in_tmp = cvCloneImage(imagePtr->getImage());
      image_vec.push_back(img_in_tmp);
    }else if (typeid(*in) == typeid(Map<int, ObjectRef>)) {
      RCPtr<Map<int, ObjectRef> > imageMapPtr = getInput(m_imageInID, count);
      Map<int, ObjectRef>::iterator it = imageMapPtr->begin();
      for (; it != imageMapPtr->end(); ++it) {
        CvfdImage& imagePtr = object_cast<CvfdImage>(it->second);
	img_in_tmp = cvCloneImage(imagePtr.getImage());
	image_vec.push_back(img_in_tmp);
      }
    }

    for(int cnt = 0; cnt < image_vec.size(); cnt++){
      IplImage* img_in;      
      img_in = cvCloneImage((IplImage*)image_vec[cnt]);
    
      // ---------------------------------------------------------------
      // main routine : start

      int i;
      IplImage *src_gray = 0;
      const char *cascade_name = "/usr/share/opencv/haarcascades/haarcascade_frontalface_default.xml";
      CvHaarClassifierCascade *cascade = 0;
      CvMemStorage *storage = 0;
      CvSeq *faces;
      static CvScalar colors[] = {
	{{0, 0, 255}}, {{0, 128, 255}},
	{{0, 255, 255}}, {{0, 255, 0}},
	{{255, 128, 0}}, {{255, 255, 0}},
	{{255, 0, 0}}, {{255, 0, 255}}
      };
    
      src_gray = cvCreateImage(cvGetSize(img_in), IPL_DEPTH_8U, 1);
    
      cascade = (CvHaarClassifierCascade *) cvLoad (cascade_name, 0, 0, 0);
    
      storage = cvCreateMemStorage (0);
      cvClearMemStorage (storage);
      cvCvtColor (img_in, src_gray, CV_BGR2GRAY);
      cvEqualizeHist (src_gray, src_gray);
    
#if ENABLE_OPENCV_2_2_0
      faces = cvHaarDetectObjects (src_gray, cascade, storage, scale_factor, min_neighbors, flags, cvSize (min_size[0], min_size[1]), cvSize (max_size[0], max_size[1])); 
#else
      faces = cvHaarDetectObjects (src_gray, cascade, storage, scale_factor, min_neighbors, flags, cvSize (min_size[0], min_size[1]));
#endif
    
      for (i = 0; i < (faces ? faces->total : 0); i++) {
	CvRect *r = (CvRect *) cvGetSeqElem (faces, i);
	CvPoint center;
	int radius;
	center.x = cvRound (r->x + r->width * 0.5);
	center.y = cvRound (r->y + r->height * 0.5);
	radius = cvRound ((r->width + r->height) * 0.25);
	// cvCircle (img_in, center, radius, colors[i % 8], 3, 8, 0);

      
	RCPtr<SrcInfo> src(new SrcInfo);
	src->roi_c[0] = center.x;
	src->roi_c[1] = center.y;
	src->roi_xy[0] = r->width;
	src->roi_xy[1] = r->height;
	src->img_size[0] = img_in->width;
	src->img_size[1] = img_in->height;
      
	sources.push_back(src);

      }

      cvReleaseImage(&img_in);
      cvReleaseImage(&src_gray);

    }

    cvReleaseImage(&img_in_tmp);

  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdhHaarDetectObjects)
};
