/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "Map.h"
#include "TimeStamp.h"
#include <csignal>

#include <../config.h>

#ifdef ENABLE_ROS

#include <ros/ros.h>
#include <ros/callback_queue.h>
#include <boost/thread.hpp>
#include <sensor_msgs/Image.h>
#include <fdopencv_msgs/ImageMap.h>
#include <fdopencv_msgs/ImageMapVal.h>
#include "HarkRosGlobals.h"
#include <sstream>
#include <string>
#include <math.h>
#include <vector>
#include <sys/time.h>
#include <queue>

using namespace std;
using namespace FD;


class callbackFuncImageStreamFromRos
{

public:
  deque<sensor_msgs::Image>           deque_image;
  deque<fdopencv_msgs::ImageMap>      deque_imagemap;
  bool enable_debug;
  int init_flag;
  int data_buffer_num;  

  callbackFuncImageStreamFromRos(int buffer, bool debug){
    data_buffer_num = buffer;
    init_flag = 1;
    enable_debug = debug;
    deque_image.resize(0);
    deque_imagemap.resize(0);
  }

  void cb_image(const sensor_msgs::Image::ConstPtr& msg)
  {
    if(enable_debug)
      ROS_INFO_STREAM("Image received [thread=" << boost::this_thread::get_id() << "]");
    sensor_msgs::Image value_image;
    value_image.header.frame_id = msg->header.frame_id;
    value_image.header.stamp = msg->header.stamp;
    value_image.height = msg->height;
    value_image.width = msg->width;
    value_image.encoding = msg->encoding;
    value_image.is_bigendian = msg->is_bigendian;
    value_image.step = msg->step;
    int datasize = msg->step * msg->height;
    value_image.data.resize(datasize);
    for(int i = 0; i < datasize; i++){
      value_image.data[i] = msg->data[i];
    }
    deque_image.push_back(value_image);
    init_flag = 0;
  }

  void cb_imagemap(const fdopencv_msgs::ImageMap::ConstPtr& msg)
  {
    if(enable_debug)
      ROS_INFO_STREAM("Image received [" << msg->count << "] [thread=" << boost::this_thread::get_id() << "]");

    fdopencv_msgs::ImageMap value_imagemap;
    value_imagemap.header.frame_id = msg->header.frame_id;
    value_imagemap.header.stamp = msg->header.stamp;
    value_imagemap.count = msg->count;
    value_imagemap.num = msg->num;
    value_imagemap.src.resize(0);
    for(int i = 0; i < msg->num; i++){
      fdopencv_msgs::ImageMapVal ImageMapValMsg;
      ImageMapValMsg.img.height = msg->src[i].img.height;
      ImageMapValMsg.img.width = msg->src[i].img.width;
      ImageMapValMsg.img.encoding = msg->src[i].img.encoding;
      ImageMapValMsg.img.is_bigendian = msg->src[i].img.is_bigendian;
      ImageMapValMsg.img.step = msg->src[i].img.step;
      int datasize = msg->src[i].img.step * msg->src[i].img.height;
      ImageMapValMsg.img.data.resize(datasize);
      for(int j = 0; j < datasize; j++){
	ImageMapValMsg.img.data[i] = msg->src[i].img.data[j];
      }
      value_imagemap.src.push_back(ImageMapValMsg);
    }
    
    deque_imagemap.push_back(value_imagemap);
    init_flag = 0;
  }

};

  
class ImageStreamFromRos;
  
DECLARE_NODE(ImageStreamFromRos)
/*Node
 *
 * @name ImageStreamFromRos
 * @category FDOPENCV_HARK:ImageIO
 * @description Subscribe an image from ROS
 * 
 * @parameter_name STREAM_SELECTION
 * @parameter_type string
 * @parameter_value Image
 * @parameter_list Image:ImageMap
 * @parameter_description Selector of stream you want to realize. All other subscription will be disabled.
 * 
 * @parameter_name TOPIC_NAME_IMAGE
 * @parameter_type string
 * @parameter_value Image
 * @parameter_description Published topic name for ROS
 * 
 * @parameter_name TOPIC_NAME_IMAGEMAP
 * @parameter_type string
 * @parameter_value ImageMap
 * @parameter_description Published topic name for ROS
 * 
 * @parameter_name ROS_LOOP_RATE
 * @parameter_type float
 * @parameter_value 1000000
 * @parameter_description This allows you to specify a frequency that you would like to loop at [Hz]. Keep this value large. (If ROS interval is shorter than HARK interval, ROS interval is overwritten.)
 * 
 * @parameter_name MSG_BUFFER_NUM
 * @parameter_type int
 * @parameter_value 100
 * @parameter_description Buffer size for a ROS subscribed message.
 * 
 * @parameter_name DATA_BUFFER_NUM
 * @parameter_type int
 * @parameter_value 100
 * @parameter_description Buffer save size for a HARK output. Too small value makes packet loss. Too large value takes large memory. 
 * 
 * @parameter_name KILL_TIME_LEN
 * @parameter_type int
 * @parameter_value 900000
 * @parameter_description Kill time length for this network [usec] (must be less than 10[sec]). Set KILL_NETWORK enable.
 *
 * @parameter_name KILL_NETWORK
 * @parameter_type bool
 * @parameter_value true
 * @parameter_list true:false
 * @parameter_description Kill whole hark network when this node does not take any message from ROS.
 *
 * @parameter_name ENABLE_DEBUG
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_description print debug message of this module in case of true.
 *
 * @output_name IMAGEOUT
 * @output_description Output image. (CvfdImage) 
 * @output_type CvfdImage
 *
 * @output_name IMAGEMAPOUT
 * @output_description Output image. (Map<int, ObjectRef>)
 * @output_type Map<int,ObjectRef>
 *
 * @output_name TIMESTAMP
 * @output_type TimeStamp
 * @output_description Time stamp of the subscribed messages
 *
 * @output_name NOT_EOF
 * @output_type bool
 * @output_description True unless End-of-File
 *
 END*/

bool exit_flag_ImageStreamFromRos = false;

void signal_handler_ImageStreamFromRos(int s)
{
  exit_flag_ImageStreamFromRos = true;
}
    
class ImageStreamFromRos : public BufferedNode {
    
  int m_imageOutID; 
  int m_imagemapOutID; 
  int timestampID;
  int eofID;

  bool bimageoutID; 
  bool bimagemapoutID; 

  string topic_name_image;
  string topic_name_imagemap;

  string stream_selection;
  bool enable_debug;
  int msg_buffer_num;  
  int data_buffer_num;
  float ros_loop_rate;
  int kill_time_len;  
  bool kill_network;
  int output_exception_flag;

  callbackFuncImageStreamFromRos *cbf;

  struct timeval tv_base;
  struct timeval tv_tmp;

private:
  ros::Subscriber _sub_image;
  ros::Subscriber _sub_imagemap;
  ros::AsyncSpinner *_snum;
    
public:
  ImageStreamFromRos(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params),
      m_imageOutID(-1), m_imagemapOutID(-1),
      bimageoutID(false), bimagemapoutID(false) 
  {

    eofID = addOutput("NOT_EOF");
    timestampID = addOutput("TIMESTAMP");
    
    topic_name_image     = object_cast<String>(parameters.get("TOPIC_NAME_IMAGE"));
    topic_name_imagemap  = object_cast<String>(parameters.get("TOPIC_NAME_IMAGEMAP"));

    stream_selection = object_cast<String>(parameters.get("STREAM_SELECTION"));
    enable_debug = dereference_cast<bool>(parameters.get("ENABLE_DEBUG"));
    ros_loop_rate = dereference_cast<float>(parameters.get("ROS_LOOP_RATE"));
    msg_buffer_num = dereference_cast<int>(parameters.get("MSG_BUFFER_NUM"));
    data_buffer_num = dereference_cast<int>(parameters.get("DATA_BUFFER_NUM"));
    kill_time_len  = dereference_cast<int>(parameters.get("KILL_TIME_LEN"));
    kill_network = dereference_cast<bool>(parameters.get("KILL_NETWORK"));
    output_exception_flag = 0;

    signal(SIGINT, signal_handler_ImageStreamFromRos);
    signal(SIGHUP, signal_handler_ImageStreamFromRos);

    cbf = new callbackFuncImageStreamFromRos(data_buffer_num, enable_debug);
    inOrder = true;
    cout << getName() << " constructor end..." << endl;

  }
      
  ~ImageStreamFromRos()
  {
  }

  virtual void initialize()
  {
    cout << getName() << " initialized..." << endl;
    if(stream_selection == "Image"){
      _sub_image    = __MasterRosNodeHandler__->subscribe<sensor_msgs::Image>(topic_name_image, msg_buffer_num, &callbackFuncImageStreamFromRos::cb_image, cbf);
      outputs[m_imageOutID].lookAhead = outputs[eofID].lookAhead = 1+max(outputs[m_imageOutID].lookAhead, outputs[eofID].lookAhead);
      outputs[m_imageOutID].lookBack  = outputs[eofID].lookBack  = 1+max(outputs[m_imageOutID].lookBack,  outputs[eofID].lookBack);
      if(m_imagemapOutID != -1) output_exception_flag = 1;
    }
    if(stream_selection == "ImageMap"){
      _sub_imagemap = __MasterRosNodeHandler__->subscribe<fdopencv_msgs::ImageMap>(topic_name_imagemap, msg_buffer_num, &callbackFuncImageStreamFromRos::cb_imagemap, cbf);
      outputs[m_imagemapOutID].lookAhead = outputs[eofID].lookAhead = 1+max(outputs[m_imagemapOutID].lookAhead, outputs[eofID].lookAhead);
      outputs[m_imagemapOutID].lookBack  = outputs[eofID].lookBack  = 1+max(outputs[m_imagemapOutID].lookBack,  outputs[eofID].lookBack);
      if(m_imageOutID != -1) output_exception_flag = 1;
    }
    this->BufferedNode::initialize();
    _snum = new ros::AsyncSpinner(1);
    _snum->start();
  }
   
  // dynamic output-port translation
  virtual int translateOutput (string outputName)
  {
    if (outputName == "IMAGEOUT") {
      if(bimageoutID){ 
	return m_imageOutID; 
      }else{ 
	bimageoutID = true; 
	return m_imageOutID = addOutput(outputName); 
      } 
    }
    if (outputName == "IMAGEMAPOUT") {
      if(bimagemapoutID){ 
	return m_imagemapOutID; 
      }else{ 
	bimagemapoutID = true; 
	return m_imagemapOutID = addOutput(outputName); 
      } 
    }
    for (unsigned int i=0; i< outputNames.size(); i++) {	
      if (outputNames[i] == outputName) {
	return i;
      }
    }      
    return addOutput(outputName);
  }
   
  void calculate(int output_id, int count, Buffer &out)
  {

    RCPtr<TimeStamp> cur_timestamp(new TimeStamp());
    (*outputs[timestampID].buffer)[count] = cur_timestamp;

    Buffer &eofBuffer = *(outputs[eofID].buffer);
    eofBuffer[count] = TrueObject;

    ros::Rate loop_rate(ros_loop_rate);

    if(output_exception_flag){
      cout << "Error : "<< getName() << " output exception -> Connected output ports are not appropriate..." << endl;
      eofBuffer[count] = FalseObject;
      return;
    }

    while(cbf->init_flag){
      if(exit_flag_ImageStreamFromRos)
	throw new NodeException(NULL, getName()+string(" Force-quit..."), __FILE__, __LINE__);
      usleep(1);
    }
    if(exit_flag_ImageStreamFromRos)
      throw new NodeException(NULL, getName()+string(" Force-quit..."), __FILE__, __LINE__);

    // Set the base time

    gettimeofday(&tv_base, NULL);
    long int usec_base = tv_base.tv_usec;
    long int sec_base  = tv_base.tv_sec % 10;
    long int sec_cnt_base = sec_base * 1000000 + usec_base;

    //----- Image message
    //////////////////////////////////////////////////////////////

    if(stream_selection == "Image"){    

      while(cbf->deque_image.size()==0){
	gettimeofday(&tv_tmp, NULL);
	long int usec_tmp = tv_tmp.tv_usec;
	long int sec_tmp  = tv_tmp.tv_sec % 10;
	if(sec_tmp < sec_base){sec_tmp += 10;}      
	long int sec_cnt_tmp = sec_tmp * 1000000 + usec_tmp;
	usleep(1);
	// If there has been no input from ROS for more than [kill_time_len]usec, kill this node
	if((sec_cnt_tmp - sec_cnt_base > kill_time_len)&&(kill_network)){
	  eofBuffer[count] = FalseObject;
	  break;
	}      
	if(exit_flag_ImageStreamFromRos)
	  throw new NodeException(NULL, getName()+string(" Force-quit..."), __FILE__, __LINE__);
      }

      // Image creation for the output
      if(cbf->deque_image.size()){

        cur_timestamp->setTime((long int)cbf->deque_image[0].header.stamp.sec, (long int)cbf->deque_image[0].header.stamp.nsec);
	
	CvSize size;
	size.width  = cbf->deque_image[0].width;
	size.height = cbf->deque_image[0].height;
	int channel = (int)(cbf->deque_image[0].step / cbf->deque_image[0].width);
	IplImage* img_in;
	img_in = cvCreateImage(size,IPL_DEPTH_8U,channel);
	for(int y=0; y < size.height; y++){
	  for(int x=0; x < size.width; x++){
	    for(int ch=0; ch < channel; ch++){
	      *(img_in->imageData+(ch + channel * x + size.width * channel * y)) = (char)cbf->deque_image[0].data[ch + channel * x + size.width * channel * y];
	    }
	  }
	}
	CvfdImage* image = new CvfdImage(img_in);
	(*(outputs[m_imageOutID].buffer))[count] = ObjectRef(image);      
	cvReleaseImage(&img_in);      
	cbf->deque_image.pop_front();
	while(cbf->deque_image.size() > cbf->data_buffer_num) cbf->deque_image.pop_front();    	
      }
      
    }

    //----- ImageMap message
    //////////////////////////////////////////////////////////////

    if(stream_selection == "ImageMap"){    

      while(cbf->deque_imagemap.size()==0){
	gettimeofday(&tv_tmp, NULL);
	long int usec_tmp = tv_tmp.tv_usec;
	long int sec_tmp  = tv_tmp.tv_sec % 10;
	if(sec_tmp < sec_base){sec_tmp += 10;}      
	long int sec_cnt_tmp = sec_tmp * 1000000 + usec_tmp;
	usleep(1);
	// If there has been no input from ROS for more than [kill_time_len]usec, kill this node
	if((sec_cnt_tmp - sec_cnt_base > kill_time_len)&&(kill_network)){
	  eofBuffer[count] = FalseObject;
	  break;
	}      
	if(exit_flag_ImageStreamFromRos)
	  throw new NodeException(NULL, getName()+string(" Force-quit..."), __FILE__, __LINE__);
      }

      RCPtr<Map<int, ObjectRef> > outputMapPtr(new Map<int, ObjectRef>);
      (*(outputs[m_imagemapOutID].buffer))[count] = outputMapPtr; 

      if(cbf->deque_imagemap.size()){

        cur_timestamp->setTime((long int)cbf->deque_imagemap[0].header.stamp.sec, (long int)cbf->deque_imagemap[0].header.stamp.nsec);

	for(int i = 0; i < cbf->deque_imagemap[0].num; i++){
	  // Image creation for the output
	  CvSize size;
	  size.width  = cbf->deque_imagemap[0].src[i].img.width;
	  size.height = cbf->deque_imagemap[0].src[i].img.height;
	  int channel = (int)(cbf->deque_imagemap[0].src[i].img.step / cbf->deque_imagemap[0].src[i].img.width);
	  IplImage* img_in;
	  img_in = cvCreateImage(size,IPL_DEPTH_8U,channel);
	  for(int y=0; y < size.height; y++){
	    for(int x=0; x < size.width; x++){
	      for(int ch=0; ch < channel; ch++){
		*(img_in->imageData+(ch + channel * x + size.width * channel * y)) = (char)cbf->deque_imagemap[0].src[i].img.data[ch + channel * x + size.width * channel * y];
	      }
	    }
	  }
	  CvfdImage* image= new CvfdImage(img_in);  
	  (*outputMapPtr)[cbf->deque_imagemap[0].src[i].id] = image;
	  cvReleaseImage(&img_in);      
	}
	cbf->deque_imagemap.pop_front();
	while(cbf->deque_imagemap.size() > cbf->data_buffer_num) cbf->deque_imagemap.pop_front();    	
      }

    }

    loop_rate.sleep();

  }
      
  NO_ORDER_NODE_SPEEDUP(ImageStreamFromRos)
};

#endif
