/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "CvfdVideo.h"

using namespace std;
using namespace FD;

class CvfdQueryFrame;
   
DECLARE_NODE(CvfdQueryFrame)
/*Node
 *
 * @name CvfdQueryFrame
 * @category FDOPENCV_ORG:Video
 * @description Take a frame in the video
 *
 * @input_name VIDEO
 * @input_description Video capturing structure
 * @input_type CvfdVideo
 *
 * @output_name IMAGE
 * @output_description The result image
 * @output_type CvfdImage
 *
 * @output_name BOOL
 * @output_description True if the video isn't finish
 * @output_type bool 
 *
 * @parameter_name INTERVAL
 * @parameter_type int
 * @parameter_value 1
 * @parameter_description Interval of file storing.
 *
 END*/
   
class CvfdQueryFrame : public BufferedNode {
      
  //Input ID
  int m_videoID;
  //Output ID
  int m_imageID;
  int m_boolID;
      
  int interval;
public:
  CvfdQueryFrame(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    //add inputs
    m_videoID = addInput("VIDEO");
         
    //add outputs
    m_imageID = addOutput("IMAGE"); 
    m_boolID = addOutput("BOOL");   

    //parameter
    interval = dereference_cast<int>(parameters.get("INTERVAL"));
    cerr << "interval :" << interval << endl;
  }
      
  void calculate(int output_id, int count, Buffer &out)
  {           

    //Read the inputs
    RCPtr<CvfdVideo> videoPtr = getInput(m_videoID,count);
         
    //Handle
    CvfdImage* image;   
         
    //Should not be released or modified
    IplImage* frame;
    bool Bool = true;
         
    // exit if interval
    if (count % interval != 0){
      image = new CvfdImage();
      (*(outputs[m_imageID].buffer))[count] = ObjectRef(image); 
      (*(outputs[m_boolID].buffer))[count] = ObjectRef(Bool::alloc(Bool));
      return;
    }

    int status = cvGetErrMode();
    cvSetErrMode( CV_ErrModeSilent );
    __BEGIN__; 
    OPENCV_CALL( frame = cvQueryFrame(videoPtr->getVideo()) );        
    __END__;
    cvSetErrMode( status );
    if( cvGetErrStatus() != CV_StsOk  )
      {
	throw new GeneralException("OPENCV - Error to take a frame: " +  CCHAR(cvErrorStr( cvGetErrStatus() )),__FILE__,__LINE__);
      }
         
    if(frame)
      {
	image = new CvfdImage(frame);            
      }
    else
      {
	Bool = false; 
	image = new CvfdImage();
      }
         
    (*(outputs[m_imageID].buffer))[count] = ObjectRef(image); 
    (*(outputs[m_boolID].buffer))[count] = ObjectRef(Bool::alloc(Bool));         
  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdQueryFrame)
};
